"Finsy is a P4Runtime framework for Python."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "0.2.0"

import sys

if sys.version_info < (3, 10):  # pragma: no cover
    raise RuntimeError("Requires Python 3.10+.")

from .controller import Controller
from .gnmiclient import gNMIClient
from .gnmipath import gNMIPath
from .p4entity import (
    P4CounterData,
    P4DigestEntry,
    P4DigestList,
    P4MeterConfig,
    P4MeterCounterData,
    P4MulticastGroupEntry,
    P4PacketIn,
    P4PacketOut,
    P4RegisterEntry,
    P4TableAction,
    P4TableEntry,
    P4TableMatch,
)
from .p4schema import P4ConfigAction, P4Schema
from .switch import Switch, SwitchEvent, SwitchOptions

__all__ = [
    "Controller",
    "P4CounterData",
    "P4DigestEntry",
    "P4DigestList",
    "P4MeterConfig",
    "P4MeterCounterData",
    "P4MulticastGroupEntry",
    "P4PacketIn",
    "P4PacketOut",
    "P4RegisterEntry",
    "P4TableAction",
    "P4TableEntry",
    "P4TableMatch",
    "P4ConfigAction",
    "P4Schema",
    "Switch",
    "SwitchEvent",
    "SwitchOptions",
    "gNMIClient",
    "gNMIPath",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import grpc
import p4.v1.p4runtime_pb2
import typing

class P4RuntimeStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Write: grpc.UnaryUnaryMultiCallable[
        p4.v1.p4runtime_pb2.WriteRequest,
        p4.v1.p4runtime_pb2.WriteResponse]
    """Update one or more P4 entities on the target."""

    Read: grpc.UnaryStreamMultiCallable[
        p4.v1.p4runtime_pb2.ReadRequest,
        p4.v1.p4runtime_pb2.ReadResponse]
    """Read one or more P4 entities from the target."""

    SetForwardingPipelineConfig: grpc.UnaryUnaryMultiCallable[
        p4.v1.p4runtime_pb2.SetForwardingPipelineConfigRequest,
        p4.v1.p4runtime_pb2.SetForwardingPipelineConfigResponse]
    """Sets the P4 forwarding-pipeline config."""

    GetForwardingPipelineConfig: grpc.UnaryUnaryMultiCallable[
        p4.v1.p4runtime_pb2.GetForwardingPipelineConfigRequest,
        p4.v1.p4runtime_pb2.GetForwardingPipelineConfigResponse]
    """Gets the current P4 forwarding-pipeline config."""

    StreamChannel: grpc.StreamStreamMultiCallable[
        p4.v1.p4runtime_pb2.StreamMessageRequest,
        p4.v1.p4runtime_pb2.StreamMessageResponse]
    """Represents the bidirectional stream between the controller and the
    switch (initiated by the controller), and is managed for the following
    purposes:
    - connection initiation through client arbitration
    - indicating switch session liveness: the session is live when switch
      sends a positive client arbitration update to the controller, and is
      considered dead when either the stream breaks or the switch sends a
      negative update for client arbitration
    - the controller sending/receiving packets to/from the switch
    - streaming of notifications from the switch
    """

    Capabilities: grpc.UnaryUnaryMultiCallable[
        p4.v1.p4runtime_pb2.CapabilitiesRequest,
        p4.v1.p4runtime_pb2.CapabilitiesResponse]


class P4RuntimeServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Write(self,
        request: p4.v1.p4runtime_pb2.WriteRequest,
        context: grpc.ServicerContext,
    ) -> p4.v1.p4runtime_pb2.WriteResponse:
        """Update one or more P4 entities on the target."""
        pass

    @abc.abstractmethod
    def Read(self,
        request: p4.v1.p4runtime_pb2.ReadRequest,
        context: grpc.ServicerContext,
    ) -> typing.Iterator[p4.v1.p4runtime_pb2.ReadResponse]:
        """Read one or more P4 entities from the target."""
        pass

    @abc.abstractmethod
    def SetForwardingPipelineConfig(self,
        request: p4.v1.p4runtime_pb2.SetForwardingPipelineConfigRequest,
        context: grpc.ServicerContext,
    ) -> p4.v1.p4runtime_pb2.SetForwardingPipelineConfigResponse:
        """Sets the P4 forwarding-pipeline config."""
        pass

    @abc.abstractmethod
    def GetForwardingPipelineConfig(self,
        request: p4.v1.p4runtime_pb2.GetForwardingPipelineConfigRequest,
        context: grpc.ServicerContext,
    ) -> p4.v1.p4runtime_pb2.GetForwardingPipelineConfigResponse:
        """Gets the current P4 forwarding-pipeline config."""
        pass

    @abc.abstractmethod
    def StreamChannel(self,
        request_iterator: typing.Iterator[p4.v1.p4runtime_pb2.StreamMessageRequest],
        context: grpc.ServicerContext,
    ) -> typing.Iterator[p4.v1.p4runtime_pb2.StreamMessageResponse]:
        """Represents the bidirectional stream between the controller and the
        switch (initiated by the controller), and is managed for the following
        purposes:
        - connection initiation through client arbitration
        - indicating switch session liveness: the session is live when switch
          sends a positive client arbitration update to the controller, and is
          considered dead when either the stream breaks or the switch sends a
          negative update for client arbitration
        - the controller sending/receiving packets to/from the switch
        - streaming of notifications from the switch
        """
        pass

    @abc.abstractmethod
    def Capabilities(self,
        request: p4.v1.p4runtime_pb2.CapabilitiesRequest,
        context: grpc.ServicerContext,
    ) -> p4.v1.p4runtime_pb2.CapabilitiesResponse: ...


def add_P4RuntimeServicer_to_server(servicer: P4RuntimeServicer, server: grpc.Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/rpc/error_details.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1egoogle/rpc/error_details.proto\x12\ngoogle.rpc\x1a\x1egoogle/protobuf/duration.proto\";\n\tRetryInfo\x12.\n\x0bretry_delay\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\"2\n\tDebugInfo\x12\x15\n\rstack_entries\x18\x01 \x03(\t\x12\x0e\n\x06\x64\x65tail\x18\x02 \x01(\t\"y\n\x0cQuotaFailure\x12\x36\n\nviolations\x18\x01 \x03(\x0b\x32\".google.rpc.QuotaFailure.Violation\x1a\x31\n\tViolation\x12\x0f\n\x07subject\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"\x93\x01\n\tErrorInfo\x12\x0e\n\x06reason\x18\x01 \x01(\t\x12\x0e\n\x06\x64omain\x18\x02 \x01(\t\x12\x35\n\x08metadata\x18\x03 \x03(\x0b\x32#.google.rpc.ErrorInfo.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x95\x01\n\x13PreconditionFailure\x12=\n\nviolations\x18\x01 \x03(\x0b\x32).google.rpc.PreconditionFailure.Violation\x1a?\n\tViolation\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"\x83\x01\n\nBadRequest\x12?\n\x10\x66ield_violations\x18\x01 \x03(\x0b\x32%.google.rpc.BadRequest.FieldViolation\x1a\x34\n\x0e\x46ieldViolation\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"7\n\x0bRequestInfo\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x14\n\x0cserving_data\x18\x02 \x01(\t\"`\n\x0cResourceInfo\x12\x15\n\rresource_type\x18\x01 \x01(\t\x12\x15\n\rresource_name\x18\x02 \x01(\t\x12\r\n\x05owner\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\"V\n\x04Help\x12$\n\x05links\x18\x01 \x03(\x0b\x32\x15.google.rpc.Help.Link\x1a(\n\x04Link\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\"3\n\x10LocalizedMessage\x12\x0e\n\x06locale\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\tBl\n\x0e\x63om.google.rpcB\x11\x45rrorDetailsProtoP\x01Z?google.golang.org/genproto/googleapis/rpc/errdetails;errdetails\xa2\x02\x03RPCb\x06proto3')



_RETRYINFO = DESCRIPTOR.message_types_by_name['RetryInfo']
_DEBUGINFO = DESCRIPTOR.message_types_by_name['DebugInfo']
_QUOTAFAILURE = DESCRIPTOR.message_types_by_name['QuotaFailure']
_QUOTAFAILURE_VIOLATION = _QUOTAFAILURE.nested_types_by_name['Violation']
_ERRORINFO = DESCRIPTOR.message_types_by_name['ErrorInfo']
_ERRORINFO_METADATAENTRY = _ERRORINFO.nested_types_by_name['MetadataEntry']
_PRECONDITIONFAILURE = DESCRIPTOR.message_types_by_name['PreconditionFailure']
_PRECONDITIONFAILURE_VIOLATION = _PRECONDITIONFAILURE.nested_types_by_name['Violation']
_BADREQUEST = DESCRIPTOR.message_types_by_name['BadRequest']
_BADREQUEST_FIELDVIOLATION = _BADREQUEST.nested_types_by_name['FieldViolation']
_REQUESTINFO = DESCRIPTOR.message_types_by_name['RequestInfo']
_RESOURCEINFO = DESCRIPTOR.message_types_by_name['ResourceInfo']
_HELP = DESCRIPTOR.message_types_by_name['Help']
_HELP_LINK = _HELP.nested_types_by_name['Link']
_LOCALIZEDMESSAGE = DESCRIPTOR.message_types_by_name['LocalizedMessage']
RetryInfo = _reflection.GeneratedProtocolMessageType('RetryInfo', (_message.Message,), {
  'DESCRIPTOR' : _RETRYINFO,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.RetryInfo)
  })
_sym_db.RegisterMessage(RetryInfo)

DebugInfo = _reflection.GeneratedProtocolMessageType('DebugInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEBUGINFO,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.DebugInfo)
  })
_sym_db.RegisterMessage(DebugInfo)

QuotaFailure = _reflection.GeneratedProtocolMessageType('QuotaFailure', (_message.Message,), {

  'Violation' : _reflection.GeneratedProtocolMessageType('Violation', (_message.Message,), {
    'DESCRIPTOR' : _QUOTAFAILURE_VIOLATION,
    '__module__' : 'google.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:google.rpc.QuotaFailure.Violation)
    })
  ,
  'DESCRIPTOR' : _QUOTAFAILURE,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.QuotaFailure)
  })
_sym_db.RegisterMessage(QuotaFailure)
_sym_db.RegisterMessage(QuotaFailure.Violation)

ErrorInfo = _reflection.GeneratedProtocolMessageType('ErrorInfo', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _ERRORINFO_METADATAENTRY,
    '__module__' : 'google.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:google.rpc.ErrorInfo.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _ERRORINFO,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.ErrorInfo)
  })
_sym_db.RegisterMessage(ErrorInfo)
_sym_db.RegisterMessage(ErrorInfo.MetadataEntry)

PreconditionFailure = _reflection.GeneratedProtocolMessageType('PreconditionFailure', (_message.Message,), {

  'Violation' : _reflection.GeneratedProtocolMessageType('Violation', (_message.Message,), {
    'DESCRIPTOR' : _PRECONDITIONFAILURE_VIOLATION,
    '__module__' : 'google.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:google.rpc.PreconditionFailure.Violation)
    })
  ,
  'DESCRIPTOR' : _PRECONDITIONFAILURE,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.PreconditionFailure)
  })
_sym_db.RegisterMessage(PreconditionFailure)
_sym_db.RegisterMessage(PreconditionFailure.Violation)

BadRequest = _reflection.GeneratedProtocolMessageType('BadRequest', (_message.Message,), {

  'FieldViolation' : _reflection.GeneratedProtocolMessageType('FieldViolation', (_message.Message,), {
    'DESCRIPTOR' : _BADREQUEST_FIELDVIOLATION,
    '__module__' : 'google.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:google.rpc.BadRequest.FieldViolation)
    })
  ,
  'DESCRIPTOR' : _BADREQUEST,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.BadRequest)
  })
_sym_db.RegisterMessage(BadRequest)
_sym_db.RegisterMessage(BadRequest.FieldViolation)

RequestInfo = _reflection.GeneratedProtocolMessageType('RequestInfo', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTINFO,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.RequestInfo)
  })
_sym_db.RegisterMessage(RequestInfo)

ResourceInfo = _reflection.GeneratedProtocolMessageType('ResourceInfo', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEINFO,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.ResourceInfo)
  })
_sym_db.RegisterMessage(ResourceInfo)

Help = _reflection.GeneratedProtocolMessageType('Help', (_message.Message,), {

  'Link' : _reflection.GeneratedProtocolMessageType('Link', (_message.Message,), {
    'DESCRIPTOR' : _HELP_LINK,
    '__module__' : 'google.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:google.rpc.Help.Link)
    })
  ,
  'DESCRIPTOR' : _HELP,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.Help)
  })
_sym_db.RegisterMessage(Help)
_sym_db.RegisterMessage(Help.Link)

LocalizedMessage = _reflection.GeneratedProtocolMessageType('LocalizedMessage', (_message.Message,), {
  'DESCRIPTOR' : _LOCALIZEDMESSAGE,
  '__module__' : 'google.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:google.rpc.LocalizedMessage)
  })
_sym_db.RegisterMessage(LocalizedMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\016com.google.rpcB\021ErrorDetailsProtoP\001Z?google.golang.org/genproto/googleapis/rpc/errdetails;errdetails\242\002\003RPC'
  _ERRORINFO_METADATAENTRY._options = None
  _ERRORINFO_METADATAENTRY._serialized_options = b'8\001'
  _RETRYINFO._serialized_start=78
  _RETRYINFO._serialized_end=137
  _DEBUGINFO._serialized_start=139
  _DEBUGINFO._serialized_end=189
  _QUOTAFAILURE._serialized_start=191
  _QUOTAFAILURE._serialized_end=312
  _QUOTAFAILURE_VIOLATION._serialized_start=263
  _QUOTAFAILURE_VIOLATION._serialized_end=312
  _ERRORINFO._serialized_start=315
  _ERRORINFO._serialized_end=462
  _ERRORINFO_METADATAENTRY._serialized_start=415
  _ERRORINFO_METADATAENTRY._serialized_end=462
  _PRECONDITIONFAILURE._serialized_start=465
  _PRECONDITIONFAILURE._serialized_end=614
  _PRECONDITIONFAILURE_VIOLATION._serialized_start=551
  _PRECONDITIONFAILURE_VIOLATION._serialized_end=614
  _BADREQUEST._serialized_start=617
  _BADREQUEST._serialized_end=748
  _BADREQUEST_FIELDVIOLATION._serialized_start=696
  _BADREQUEST_FIELDVIOLATION._serialized_end=748
  _REQUESTINFO._serialized_start=750
  _REQUESTINFO._serialized_end=805
  _RESOURCEINFO._serialized_start=807
  _RESOURCEINFO._serialized_end=903
  _HELP._serialized_start=905
  _HELP._serialized_end=991
  _HELP_LINK._serialized_start=951
  _HELP_LINK._serialized_end=991
  _LOCALIZEDMESSAGE._serialized_start=993
  _LOCALIZEDMESSAGE._serialized_end=1044
# @@protoc_insertion_point(module_scope)



"Implements the PortList and Port classes."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from dataclasses import dataclass

import finsy.switch as _sw  # circular import
from finsy.gnmiclient import gNMIClient, gNMIPath, gNMISubscription, gNMIUpdate
from finsy.log import LOGGER, TRACE

# FIXME: Use GNMI id or ifIndex?
_ifIndex = gNMIPath("interfaces/interface[name=*]/state/ifindex")
_ifOperStatus = gNMIPath("interfaces/interface[name=*]/state/oper-status")


class OperStatus(enum.Enum):
    "gNMI values for ifOperStatus"

    DORMANT = "DORMANT"
    LOWER_LAYER_DOWN = "LOWER_LAYER_DOWN"
    UNKNOWN = "UNKNOWN"
    TESTING = "TESTING"
    UP = "UP"
    DOWN = "DOWN"
    NOT_PRESENT = "NOT_PRESENT"


@dataclass
class Port:
    "Represents a switch port."

    id: int
    name: str
    oper_status: OperStatus = OperStatus.UNKNOWN

    @property
    def up(self) -> bool:
        "Return true if port is basically up."
        return self.oper_status == OperStatus.UP


class PortList:
    "Represents a list of switch ports."

    _ports: dict[str, Port]
    _subscription: gNMISubscription | None = None

    def __init__(self):
        self._ports = {}

    def __getitem__(self, key: str):
        "Retrieve interface by ID."
        return self._ports[key]

    def __len__(self):
        return len(self._ports)

    def __iter__(self):
        return iter(self._ports.values())

    @TRACE
    async def subscribe(self, client: gNMIClient):
        assert self._subscription is None

        self._ports = await self._get_ports(client)
        self._subscription = await self._get_subscription(client)

    @TRACE
    async def update(self, switch: "_sw.Switch | None" = None):
        assert self._subscription is not None

        async for update in self._subscription.updates():
            self._update(update, switch)

    def close(self):
        if self._subscription is not None:
            self._subscription.cancel()
            self._subscription = None
            self._ports = {}

    async def _get_ports(self, client: gNMIClient) -> dict[str, Port]:
        "Retrieve ID and name of each port."
        ports: dict[str, Port] = {}

        result = await client.get(_ifIndex)
        for update in result:
            path = update.path
            assert path.last == _ifIndex.last

            port = Port(update.value, path["name"])
            ports[port.name] = port

        return ports

    async def _get_subscription(self, client: gNMIClient):
        sub = client.subscribe()

        # Subscribe to change notifications.
        for port in self._ports.values():
            sub.on_change(_ifOperStatus.set(name=port.name))

        # Synchronize initial settings for ports.
        async for update in sub.synchronize():
            self._update(update, None)

        return sub

    def _update(self, update: gNMIUpdate, switch: "_sw.Switch | None"):
        path = update.path
        name = path["name"]

        match path.last:
            case _ifOperStatus.last:
                status = OperStatus(update.value)
                self._update_port(name, status, switch)
            case _:
                LOGGER.warning(f"PortList: unknown gNMI path: {path}")

    def _update_port(self, name: str, status: OperStatus, switch: "_sw.Switch | None"):
        port = self._ports[name]

        prev_up = port.up
        port.oper_status = status
        curr_up = port.up

        if switch is not None and curr_up != prev_up:
            if curr_up:
                switch.ee.emit(_sw.SwitchEvent.PORT_UP, switch, port)
            else:
                switch.ee.emit(_sw.SwitchEvent.PORT_DOWN, switch, port)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p4/config/v1/p4info.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from p4.config.v1 import p4types_pb2 as p4_dot_config_dot_v1_dot_p4types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19p4/config/v1/p4info.proto\x12\x0cp4.config.v1\x1a\x19google/protobuf/any.proto\x1a\x1ap4/config/v1/p4types.proto\"\x88\x05\n\x06P4Info\x12\'\n\x08pkg_info\x18\x01 \x01(\x0b\x32\x15.p4.config.v1.PkgInfo\x12#\n\x06tables\x18\x02 \x03(\x0b\x32\x13.p4.config.v1.Table\x12%\n\x07\x61\x63tions\x18\x03 \x03(\x0b\x32\x14.p4.config.v1.Action\x12\x34\n\x0f\x61\x63tion_profiles\x18\x04 \x03(\x0b\x32\x1b.p4.config.v1.ActionProfile\x12\'\n\x08\x63ounters\x18\x05 \x03(\x0b\x32\x15.p4.config.v1.Counter\x12\x34\n\x0f\x64irect_counters\x18\x06 \x03(\x0b\x32\x1b.p4.config.v1.DirectCounter\x12#\n\x06meters\x18\x07 \x03(\x0b\x32\x13.p4.config.v1.Meter\x12\x30\n\rdirect_meters\x18\x08 \x03(\x0b\x32\x19.p4.config.v1.DirectMeter\x12J\n\x1a\x63ontroller_packet_metadata\x18\t \x03(\x0b\x32&.p4.config.v1.ControllerPacketMetadata\x12*\n\nvalue_sets\x18\n \x03(\x0b\x32\x16.p4.config.v1.ValueSet\x12)\n\tregisters\x18\x0b \x03(\x0b\x32\x16.p4.config.v1.Register\x12%\n\x07\x64igests\x18\x0c \x03(\x0b\x32\x14.p4.config.v1.Digest\x12%\n\x07\x65xterns\x18\x64 \x03(\x0b\x32\x14.p4.config.v1.Extern\x12,\n\ttype_info\x18\xc8\x01 \x01(\x0b\x32\x18.p4.config.v1.P4TypeInfo\"3\n\rDocumentation\x12\r\n\x05\x62rief\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"\xa9\x02\n\x07PkgInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12(\n\x03\x64oc\x18\x03 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12\x13\n\x0b\x61nnotations\x18\x04 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\n \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x0c\n\x04\x61rch\x18\x05 \x01(\t\x12\x14\n\x0corganization\x18\x06 \x01(\t\x12\x0f\n\x07\x63ontact\x18\x07 \x01(\t\x12\x0b\n\x03url\x18\x08 \x01(\t\x12\x42\n\x16structured_annotations\x18\t \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"\x87\x02\n\x05P4Ids\"\xfd\x01\n\x06Prefix\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TION\x10\x01\x12\t\n\x05TABLE\x10\x02\x12\r\n\tVALUE_SET\x10\x03\x12\x15\n\x11\x43ONTROLLER_HEADER\x10\x04\x12\x15\n\x11PSA_EXTERNS_START\x10\x10\x12\x12\n\x0e\x41\x43TION_PROFILE\x10\x11\x12\x0b\n\x07\x43OUNTER\x10\x12\x12\x12\n\x0e\x44IRECT_COUNTER\x10\x13\x12\t\n\x05METER\x10\x14\x12\x10\n\x0c\x44IRECT_METER\x10\x15\x12\x0c\n\x08REGISTER\x10\x16\x12\n\n\x06\x44IGEST\x10\x17\x12\x18\n\x13OTHER_EXTERNS_START\x10\x80\x01\x12\x08\n\x03MAX\x10\xff\x01\"\xf2\x01\n\x08Preamble\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x04 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x07 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12(\n\x03\x64oc\x18\x05 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12\x42\n\x16structured_annotations\x18\x06 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"k\n\x06\x45xtern\x12\x16\n\x0e\x65xtern_type_id\x18\x01 \x01(\r\x12\x18\n\x10\x65xtern_type_name\x18\x02 \x01(\t\x12/\n\tinstances\x18\x03 \x03(\x0b\x32\x1c.p4.config.v1.ExternInstance\"^\n\x0e\x45xternInstance\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\"\n\x04info\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"\xdc\x03\n\nMatchField\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\n \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12\x38\n\nmatch_type\x18\x05 \x01(\x0e\x32\".p4.config.v1.MatchField.MatchTypeH\x00\x12\x1a\n\x10other_match_type\x18\x07 \x01(\tH\x00\x12(\n\x03\x64oc\x18\x06 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12,\n\ttype_name\x18\x08 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\t \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"V\n\tMatchType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x45XACT\x10\x02\x12\x07\n\x03LPM\x10\x03\x12\x0b\n\x07TERNARY\x10\x04\x12\t\n\x05RANGE\x10\x05\x12\x0c\n\x08OPTIONAL\x10\x06\x42\x07\n\x05match\"\xc1\x03\n\x05Table\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12.\n\x0cmatch_fields\x18\x02 \x03(\x0b\x32\x18.p4.config.v1.MatchField\x12,\n\x0b\x61\x63tion_refs\x18\x03 \x03(\x0b\x32\x17.p4.config.v1.ActionRef\x12\x1f\n\x17\x63onst_default_action_id\x18\x04 \x01(\r\x12\x19\n\x11implementation_id\x18\x06 \x01(\r\x12\x1b\n\x13\x64irect_resource_ids\x18\x07 \x03(\r\x12\x0c\n\x04size\x18\x08 \x01(\x03\x12\x46\n\x15idle_timeout_behavior\x18\t \x01(\x0e\x32\'.p4.config.v1.Table.IdleTimeoutBehavior\x12\x16\n\x0eis_const_table\x18\n \x01(\x08\x12.\n\x10other_properties\x18\x64 \x01(\x0b\x32\x14.google.protobuf.Any\"9\n\x13IdleTimeoutBehavior\x12\x0e\n\nNO_TIMEOUT\x10\x00\x12\x12\n\x0eNOTIFY_CONTROL\x10\x01\"\x9c\x02\n\tActionRef\x12\n\n\x02id\x18\x01 \x01(\r\x12,\n\x05scope\x18\x03 \x01(\x0e\x32\x1d.p4.config.v1.ActionRef.Scope\x12\x13\n\x0b\x61nnotations\x18\x02 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x05 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x42\n\x16structured_annotations\x18\x04 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"@\n\x05Scope\x12\x15\n\x11TABLE_AND_DEFAULT\x10\x00\x12\x0e\n\nTABLE_ONLY\x10\x01\x12\x10\n\x0c\x44\x45\x46\x41ULT_ONLY\x10\x02\"\x81\x03\n\x06\x41\x63tion\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12*\n\x06params\x18\x02 \x03(\x0b\x32\x1a.p4.config.v1.Action.Param\x1a\xa0\x02\n\x05Param\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x08 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12(\n\x03\x64oc\x18\x05 \x01(\x0b\x32\x1b.p4.config.v1.Documentation\x12,\n\ttype_name\x18\x06 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\x07 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"\x9e\x02\n\rActionProfile\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\x11\n\ttable_ids\x18\x02 \x03(\r\x12\x15\n\rwith_selector\x18\x03 \x01(\x08\x12\x0c\n\x04size\x18\x04 \x01(\x03\x12\x16\n\x0emax_group_size\x18\x05 \x01(\x05\x12R\n\x17selector_size_semantics\x18\x06 \x01(\x0e\x32\x31.p4.config.v1.ActionProfile.SelectorSizeSemantics\"?\n\x15SelectorSizeSemantics\x12\x12\n\x0eSUM_OF_WEIGHTS\x10\x00\x12\x12\n\x0eSUM_OF_MEMBERS\x10\x01\"v\n\x0b\x43ounterSpec\x12,\n\x04unit\x18\x01 \x01(\x0e\x32\x1e.p4.config.v1.CounterSpec.Unit\"9\n\x04Unit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\x0b\n\x07PACKETS\x10\x02\x12\x08\n\x04\x42OTH\x10\x03\"\x9e\x01\n\x07\x43ounter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x04spec\x18\x02 \x01(\x0b\x32\x19.p4.config.v1.CounterSpec\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"{\n\rDirectCounter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x04spec\x18\x02 \x01(\x0b\x32\x19.p4.config.v1.CounterSpec\x12\x17\n\x0f\x64irect_table_id\x18\x03 \x01(\r\"h\n\tMeterSpec\x12*\n\x04unit\x18\x01 \x01(\x0e\x32\x1c.p4.config.v1.MeterSpec.Unit\"/\n\x04Unit\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\x0b\n\x07PACKETS\x10\x02\"\x9a\x01\n\x05Meter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12%\n\x04spec\x18\x02 \x01(\x0b\x32\x17.p4.config.v1.MeterSpec\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"w\n\x0b\x44irectMeter\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12%\n\x04spec\x18\x02 \x01(\x0b\x32\x17.p4.config.v1.MeterSpec\x12\x17\n\x0f\x64irect_table_id\x18\x03 \x01(\r\"\x83\x03\n\x18\x43ontrollerPacketMetadata\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\x41\n\x08metadata\x18\x02 \x03(\x0b\x32/.p4.config.v1.ControllerPacketMetadata.Metadata\x1a\xf9\x01\n\x08Metadata\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x61nnotations\x18\x03 \x03(\t\x12:\n\x14\x61nnotation_locations\x18\x07 \x03(\x0b\x32\x1c.p4.config.v1.SourceLocation\x12\x10\n\x08\x62itwidth\x18\x04 \x01(\x05\x12,\n\ttype_name\x18\x05 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\x12\x42\n\x16structured_annotations\x18\x06 \x03(\x0b\x32\".p4.config.v1.StructuredAnnotation\"k\n\x08ValueSet\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12\'\n\x05match\x18\x02 \x03(\x0b\x32\x18.p4.config.v1.MatchField\x12\x0c\n\x04size\x18\x03 \x01(\x05\"\xa7\x01\n\x08Register\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12/\n\ttype_spec\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpec\x12\x0c\n\x04size\x18\x03 \x01(\x05\x12\x32\n\x0findex_type_name\x18\x04 \x01(\x0b\x32\x19.p4.config.v1.P4NamedType\"c\n\x06\x44igest\x12(\n\x08preamble\x18\x01 \x01(\x0b\x32\x16.p4.config.v1.Preamble\x12/\n\ttype_spec\x18\x02 \x01(\x0b\x32\x1c.p4.config.v1.P4DataTypeSpecB-Z+github.com/p4lang/p4runtime/go/p4/config/v1b\x06proto3')



_P4INFO = DESCRIPTOR.message_types_by_name['P4Info']
_DOCUMENTATION = DESCRIPTOR.message_types_by_name['Documentation']
_PKGINFO = DESCRIPTOR.message_types_by_name['PkgInfo']
_P4IDS = DESCRIPTOR.message_types_by_name['P4Ids']
_PREAMBLE = DESCRIPTOR.message_types_by_name['Preamble']
_EXTERN = DESCRIPTOR.message_types_by_name['Extern']
_EXTERNINSTANCE = DESCRIPTOR.message_types_by_name['ExternInstance']
_MATCHFIELD = DESCRIPTOR.message_types_by_name['MatchField']
_TABLE = DESCRIPTOR.message_types_by_name['Table']
_ACTIONREF = DESCRIPTOR.message_types_by_name['ActionRef']
_ACTION = DESCRIPTOR.message_types_by_name['Action']
_ACTION_PARAM = _ACTION.nested_types_by_name['Param']
_ACTIONPROFILE = DESCRIPTOR.message_types_by_name['ActionProfile']
_COUNTERSPEC = DESCRIPTOR.message_types_by_name['CounterSpec']
_COUNTER = DESCRIPTOR.message_types_by_name['Counter']
_DIRECTCOUNTER = DESCRIPTOR.message_types_by_name['DirectCounter']
_METERSPEC = DESCRIPTOR.message_types_by_name['MeterSpec']
_METER = DESCRIPTOR.message_types_by_name['Meter']
_DIRECTMETER = DESCRIPTOR.message_types_by_name['DirectMeter']
_CONTROLLERPACKETMETADATA = DESCRIPTOR.message_types_by_name['ControllerPacketMetadata']
_CONTROLLERPACKETMETADATA_METADATA = _CONTROLLERPACKETMETADATA.nested_types_by_name['Metadata']
_VALUESET = DESCRIPTOR.message_types_by_name['ValueSet']
_REGISTER = DESCRIPTOR.message_types_by_name['Register']
_DIGEST = DESCRIPTOR.message_types_by_name['Digest']
_P4IDS_PREFIX = _P4IDS.enum_types_by_name['Prefix']
_MATCHFIELD_MATCHTYPE = _MATCHFIELD.enum_types_by_name['MatchType']
_TABLE_IDLETIMEOUTBEHAVIOR = _TABLE.enum_types_by_name['IdleTimeoutBehavior']
_ACTIONREF_SCOPE = _ACTIONREF.enum_types_by_name['Scope']
_ACTIONPROFILE_SELECTORSIZESEMANTICS = _ACTIONPROFILE.enum_types_by_name['SelectorSizeSemantics']
_COUNTERSPEC_UNIT = _COUNTERSPEC.enum_types_by_name['Unit']
_METERSPEC_UNIT = _METERSPEC.enum_types_by_name['Unit']
P4Info = _reflection.GeneratedProtocolMessageType('P4Info', (_message.Message,), {
  'DESCRIPTOR' : _P4INFO,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.P4Info)
  })
_sym_db.RegisterMessage(P4Info)

Documentation = _reflection.GeneratedProtocolMessageType('Documentation', (_message.Message,), {
  'DESCRIPTOR' : _DOCUMENTATION,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Documentation)
  })
_sym_db.RegisterMessage(Documentation)

PkgInfo = _reflection.GeneratedProtocolMessageType('PkgInfo', (_message.Message,), {
  'DESCRIPTOR' : _PKGINFO,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.PkgInfo)
  })
_sym_db.RegisterMessage(PkgInfo)

P4Ids = _reflection.GeneratedProtocolMessageType('P4Ids', (_message.Message,), {
  'DESCRIPTOR' : _P4IDS,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.P4Ids)
  })
_sym_db.RegisterMessage(P4Ids)

Preamble = _reflection.GeneratedProtocolMessageType('Preamble', (_message.Message,), {
  'DESCRIPTOR' : _PREAMBLE,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Preamble)
  })
_sym_db.RegisterMessage(Preamble)

Extern = _reflection.GeneratedProtocolMessageType('Extern', (_message.Message,), {
  'DESCRIPTOR' : _EXTERN,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Extern)
  })
_sym_db.RegisterMessage(Extern)

ExternInstance = _reflection.GeneratedProtocolMessageType('ExternInstance', (_message.Message,), {
  'DESCRIPTOR' : _EXTERNINSTANCE,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.ExternInstance)
  })
_sym_db.RegisterMessage(ExternInstance)

MatchField = _reflection.GeneratedProtocolMessageType('MatchField', (_message.Message,), {
  'DESCRIPTOR' : _MATCHFIELD,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.MatchField)
  })
_sym_db.RegisterMessage(MatchField)

Table = _reflection.GeneratedProtocolMessageType('Table', (_message.Message,), {
  'DESCRIPTOR' : _TABLE,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Table)
  })
_sym_db.RegisterMessage(Table)

ActionRef = _reflection.GeneratedProtocolMessageType('ActionRef', (_message.Message,), {
  'DESCRIPTOR' : _ACTIONREF,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.ActionRef)
  })
_sym_db.RegisterMessage(ActionRef)

Action = _reflection.GeneratedProtocolMessageType('Action', (_message.Message,), {

  'Param' : _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), {
    'DESCRIPTOR' : _ACTION_PARAM,
    '__module__' : 'p4.config.v1.p4info_pb2'
    # @@protoc_insertion_point(class_scope:p4.config.v1.Action.Param)
    })
  ,
  'DESCRIPTOR' : _ACTION,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Action)
  })
_sym_db.RegisterMessage(Action)
_sym_db.RegisterMessage(Action.Param)

ActionProfile = _reflection.GeneratedProtocolMessageType('ActionProfile', (_message.Message,), {
  'DESCRIPTOR' : _ACTIONPROFILE,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.ActionProfile)
  })
_sym_db.RegisterMessage(ActionProfile)

CounterSpec = _reflection.GeneratedProtocolMessageType('CounterSpec', (_message.Message,), {
  'DESCRIPTOR' : _COUNTERSPEC,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.CounterSpec)
  })
_sym_db.RegisterMessage(CounterSpec)

Counter = _reflection.GeneratedProtocolMessageType('Counter', (_message.Message,), {
  'DESCRIPTOR' : _COUNTER,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Counter)
  })
_sym_db.RegisterMessage(Counter)

DirectCounter = _reflection.GeneratedProtocolMessageType('DirectCounter', (_message.Message,), {
  'DESCRIPTOR' : _DIRECTCOUNTER,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.DirectCounter)
  })
_sym_db.RegisterMessage(DirectCounter)

MeterSpec = _reflection.GeneratedProtocolMessageType('MeterSpec', (_message.Message,), {
  'DESCRIPTOR' : _METERSPEC,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.MeterSpec)
  })
_sym_db.RegisterMessage(MeterSpec)

Meter = _reflection.GeneratedProtocolMessageType('Meter', (_message.Message,), {
  'DESCRIPTOR' : _METER,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Meter)
  })
_sym_db.RegisterMessage(Meter)

DirectMeter = _reflection.GeneratedProtocolMessageType('DirectMeter', (_message.Message,), {
  'DESCRIPTOR' : _DIRECTMETER,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.DirectMeter)
  })
_sym_db.RegisterMessage(DirectMeter)

ControllerPacketMetadata = _reflection.GeneratedProtocolMessageType('ControllerPacketMetadata', (_message.Message,), {

  'Metadata' : _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
    'DESCRIPTOR' : _CONTROLLERPACKETMETADATA_METADATA,
    '__module__' : 'p4.config.v1.p4info_pb2'
    # @@protoc_insertion_point(class_scope:p4.config.v1.ControllerPacketMetadata.Metadata)
    })
  ,
  'DESCRIPTOR' : _CONTROLLERPACKETMETADATA,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.ControllerPacketMetadata)
  })
_sym_db.RegisterMessage(ControllerPacketMetadata)
_sym_db.RegisterMessage(ControllerPacketMetadata.Metadata)

ValueSet = _reflection.GeneratedProtocolMessageType('ValueSet', (_message.Message,), {
  'DESCRIPTOR' : _VALUESET,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.ValueSet)
  })
_sym_db.RegisterMessage(ValueSet)

Register = _reflection.GeneratedProtocolMessageType('Register', (_message.Message,), {
  'DESCRIPTOR' : _REGISTER,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Register)
  })
_sym_db.RegisterMessage(Register)

Digest = _reflection.GeneratedProtocolMessageType('Digest', (_message.Message,), {
  'DESCRIPTOR' : _DIGEST,
  '__module__' : 'p4.config.v1.p4info_pb2'
  # @@protoc_insertion_point(class_scope:p4.config.v1.Digest)
  })
_sym_db.RegisterMessage(Digest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z+github.com/p4lang/p4runtime/go/p4/config/v1'
  _P4INFO._serialized_start=99
  _P4INFO._serialized_end=747
  _DOCUMENTATION._serialized_start=749
  _DOCUMENTATION._serialized_end=800
  _PKGINFO._serialized_start=803
  _PKGINFO._serialized_end=1100
  _P4IDS._serialized_start=1103
  _P4IDS._serialized_end=1366
  _P4IDS_PREFIX._serialized_start=1113
  _P4IDS_PREFIX._serialized_end=1366
  _PREAMBLE._serialized_start=1369
  _PREAMBLE._serialized_end=1611
  _EXTERN._serialized_start=1613
  _EXTERN._serialized_end=1720
  _EXTERNINSTANCE._serialized_start=1722
  _EXTERNINSTANCE._serialized_end=1816
  _MATCHFIELD._serialized_start=1819
  _MATCHFIELD._serialized_end=2295
  _MATCHFIELD_MATCHTYPE._serialized_start=2200
  _MATCHFIELD_MATCHTYPE._serialized_end=2286
  _TABLE._serialized_start=2298
  _TABLE._serialized_end=2747
  _TABLE_IDLETIMEOUTBEHAVIOR._serialized_start=2690
  _TABLE_IDLETIMEOUTBEHAVIOR._serialized_end=2747
  _ACTIONREF._serialized_start=2750
  _ACTIONREF._serialized_end=3034
  _ACTIONREF_SCOPE._serialized_start=2970
  _ACTIONREF_SCOPE._serialized_end=3034
  _ACTION._serialized_start=3037
  _ACTION._serialized_end=3422
  _ACTION_PARAM._serialized_start=3134
  _ACTION_PARAM._serialized_end=3422
  _ACTIONPROFILE._serialized_start=3425
  _ACTIONPROFILE._serialized_end=3711
  _ACTIONPROFILE_SELECTORSIZESEMANTICS._serialized_start=3648
  _ACTIONPROFILE_SELECTORSIZESEMANTICS._serialized_end=3711
  _COUNTERSPEC._serialized_start=3713
  _COUNTERSPEC._serialized_end=3831
  _COUNTERSPEC_UNIT._serialized_start=3774
  _COUNTERSPEC_UNIT._serialized_end=3831
  _COUNTER._serialized_start=3834
  _COUNTER._serialized_end=3992
  _DIRECTCOUNTER._serialized_start=3994
  _DIRECTCOUNTER._serialized_end=4117
  _METERSPEC._serialized_start=4119
  _METERSPEC._serialized_end=4223
  _METERSPEC_UNIT._serialized_start=3774
  _METERSPEC_UNIT._serialized_end=3821
  _METER._serialized_start=4226
  _METER._serialized_end=4380
  _DIRECTMETER._serialized_start=4382
  _DIRECTMETER._serialized_end=4501
  _CONTROLLERPACKETMETADATA._serialized_start=4504
  _CONTROLLERPACKETMETADATA._serialized_end=4891
  _CONTROLLERPACKETMETADATA_METADATA._serialized_start=4642
  _CONTROLLERPACKETMETADATA_METADATA._serialized_end=4891
  _VALUESET._serialized_start=4893
  _VALUESET._serialized_end=5000
  _REGISTER._serialized_start=5003
  _REGISTER._serialized_end=5170
  _DIGEST._serialized_start=5172
  _DIGEST._serialized_end=5271
# @@protoc_insertion_point(module_scope)



"Utility functions for opening a GRPC channel."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from dataclasses import dataclass
from typing import Sequence

import grpc  # pyright: ignore[reportMissingTypeStubs]
from typing_extensions import Self

from finsy.log import LOGGER
from finsy.proto import rpc_code

# `grpc.aio.EOF` is not typed.
GRPC_EOF: object = grpc.aio.EOF  # pyright: ignore[reportUnknownMemberType]


class _EnumBase(enum.IntEnum):
    "Base class for GRPC/Protobuf Enum wrapper classes."

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


class GRPCStatusCode(_EnumBase):
    "IntEnum equivalent to `grpc.StatusCode`."

    OK = rpc_code.OK
    CANCELLED = rpc_code.CANCELLED
    UNKNOWN = rpc_code.UNKNOWN
    FAILED_PRECONDITION = rpc_code.FAILED_PRECONDITION
    INVALID_ARGUMENT = rpc_code.INVALID_ARGUMENT
    DEADLINE_EXCEEDED = rpc_code.DEADLINE_EXCEEDED
    NOT_FOUND = rpc_code.NOT_FOUND
    ALREADY_EXISTS = rpc_code.ALREADY_EXISTS
    PERMISSION_DENIED = rpc_code.PERMISSION_DENIED
    UNAUTHENTICATED = rpc_code.UNAUTHENTICATED
    RESOURCE_EXHAUSTED = rpc_code.RESOURCE_EXHAUSTED
    ABORTED = rpc_code.ABORTED
    OUT_OF_RANGE = rpc_code.OUT_OF_RANGE
    UNIMPLEMENTED = rpc_code.UNIMPLEMENTED
    INTERNAL = rpc_code.INTERNAL
    UNAVAILABLE = rpc_code.UNAVAILABLE
    DATA_LOSS = rpc_code.DATA_LOSS

    @classmethod
    def from_status_code(cls, val: grpc.StatusCode) -> Self:
        "Create corresponding GRPCStatusCode from a grpc.StatusCode object."
        assert isinstance(val, grpc.StatusCode)
        return GRPCStatusCode(val.value[0])

    @staticmethod
    def _validate_enum() -> None:
        for value in grpc.StatusCode:
            assert GRPCStatusCode[value.name].value == value.value[0], value.name


# Check GRPCStatusCode against grpc.StatusCode.
GRPCStatusCode._validate_enum()  # pyright: ignore[reportPrivateUsage]


class GRPCArg(str, enum.Enum):
    """Names for GRPC channel arguments.

    Reference:
    https://grpc.github.io/grpc/core/group__grpc__arg__keys.html
    """

    MAX_METADATA_SIZE = "grpc.max_metadata_size"
    MAX_RECONNECT_BACKOFF_MS = "grpc.max_reconnect_backoff_ms"


@dataclass
class GRPCOptions:
    "GRPC channel arguments."

    max_metadata_size: int | None = None
    max_reconnect_backoff_ms: int | None = None

    def args(self) -> Sequence[tuple[str, int]]:
        results: list[tuple[str, int]] = []

        if self.max_metadata_size is not None:
            results.append((GRPCArg.MAX_METADATA_SIZE, self.max_metadata_size))

        if self.max_reconnect_backoff_ms is not None:
            results.append(
                (GRPCArg.MAX_RECONNECT_BACKOFF_MS, self.max_reconnect_backoff_ms)
            )

        return results


def grpc_channel(
    address: str,
    *,
    credentials: grpc.ChannelCredentials | None = None,
    options: GRPCOptions | None = None,
    client_type: str = "GRPC",
) -> grpc.aio.Channel:
    "Create a GRPC AIO channel."

    args = options.args() if options else None

    if credentials is None:
        LOGGER.debug("%s: create INSECURE channel %r", client_type, address)
        channel = grpc.aio.insecure_channel(
            address,
            options=args,
        )
    else:
        LOGGER.debug("%s: create secure channel %r", client_type, address)
        channel = grpc.aio.secure_channel(
            address,
            credentials=credentials,
            options=args,
        )

    return channel



"Implements the Switch class."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import dataclasses
import enum
import re
import time
from contextlib import asynccontextmanager
from pathlib import Path
from types import TracebackType
from typing import Any, AsyncIterator, Callable, Coroutine, SupportsBytes, TypeVar

import grpc  # pyright: ignore[reportMissingTypeStubs]
import pyee

from finsy import p4entity
from finsy.arbitrator import Arbitrator
from finsy.gnmiclient import gNMIClient, gNMIClientError
from finsy.log import LOGGER, TRACE
from finsy.p4client import P4Client, P4ClientError
from finsy.p4schema import P4ConfigAction, P4ConfigResponseType, P4Schema, P4UpdateType
from finsy.ports import PortList
from finsy.proto import p4r
from finsy.util import CountdownFuture

# Maximum size of queues used for PacketIn, DigestList, etc.

_DEFAULT_QUEUE_SIZE = 50

# If switch fails to run for longer than _FAIL_MIN_TIME_SECS, wait
# _FAIL_THROTTLE_SECS before trying again.

_FAIL_MIN_TIME_SECS = 2.0
_FAIL_THROTTLE_SECS = 10.0

_ApiVersion = tuple[int, int, int]

_T = TypeVar("_T")


@dataclasses.dataclass(frozen=True)
class SwitchOptions:
    """`SwitchOptions` manages the configuration options for a `Switch`.

    Each `SwitchOptions` object is immutable and may be shared by multiple
    switches. Although each object is immutable, you can use function call
    syntax to change one or more properties and construct a new object.

    ```
    opts = SwitchOptions(device_id=5)
    new_opts = opts(device_id=6)
    ```
    """

    p4info: Path | None = None
    p4blob: Path | SupportsBytes | None = None
    device_id: int = 1
    initial_election_id: int = 10
    credentials: grpc.ChannelCredentials | None = None
    ready_handler: Callable[["Switch"], Coroutine[Any, Any, None]] | None = None

    def __call__(self, **kwds: Any):
        return dataclasses.replace(self, **kwds)


class Switch:
    """`Switch` manages a P4Runtime Switch.

    A `Switch` is constructed with a `name`, `address` and an optional
    `SwitchOptions` configuration.
    """

    _name: str
    _address: str
    _options: SwitchOptions
    _p4client: P4Client | None
    _p4schema: P4Schema
    _tasks: "SwitchTasks | None"
    _queues: dict[str, asyncio.Queue[Any]]
    _arbitrator: "Arbitrator"
    _gnmi_client: gNMIClient | None
    _ports: PortList
    _is_channel_up: bool = False
    _api_version: _ApiVersion = (1, 0, 0)

    control_task: asyncio.Task[Any] | None = None
    "Used by Controller to track switch's main task."

    ee: "SwitchEmitter"
    "Event emitter."

    def __init__(
        self,
        name: str,
        address: str,
        options: SwitchOptions | None = None,
    ) -> None:
        if options is None:
            options = SwitchOptions()

        self._name = name
        self._address = address
        self._options = options
        self._p4client = None
        self._p4schema = P4Schema(options.p4info, options.p4blob)
        self._tasks = None
        self._queues = {}
        self._arbitrator = Arbitrator(options.initial_election_id)
        self._gnmi_client = None
        self._ports = PortList()
        self.ee = SwitchEmitter(self)

    @property
    def name(self) -> str:
        return self._name

    @property
    def options(self) -> SwitchOptions:
        return self._options

    @options.setter
    def options(self, opts: SwitchOptions):
        if self._p4client is not None:
            raise RuntimeError("Cannot change switch options while client is open.")

        self._options = opts
        self._p4schema = P4Schema(opts.p4info, opts.p4blob)
        self._arbitrator = Arbitrator(opts.initial_election_id)

    @property
    def device_id(self) -> int:
        return self._options.device_id

    @property
    def is_primary(self) -> bool:
        return self._arbitrator.is_primary

    @property
    def primary_id(self) -> int:
        return self._arbitrator.primary_id

    @property
    def election_id(self) -> int:
        return self._arbitrator.election_id

    @property
    def p4info(self) -> P4Schema:
        return self._p4schema

    @property
    def gnmi_client(self) -> gNMIClient | None:
        return self._gnmi_client

    @property
    def ports(self) -> PortList:
        return self._ports

    @property
    def api_version(self) -> _ApiVersion:
        return self._api_version

    @property
    def api_version_string(self) -> str:
        return ".".join(map(str, self.api_version))

    def packet_iterator(
        self,
        *,
        size: int = _DEFAULT_QUEUE_SIZE,
    ) -> AsyncIterator["p4entity.P4PacketIn"]:
        return self._queue_iter("packet", size)

    def digest_iterator(
        self,
        *,
        size: int = _DEFAULT_QUEUE_SIZE,
    ) -> AsyncIterator["p4entity.P4DigestList"]:
        return self._queue_iter("digest", size)

    async def _queue_iter(self, name: str, size: int) -> AsyncIterator[Any]:
        if name in self._queues:
            raise RuntimeError(f"iterator {name!r} already open")

        LOGGER.debug("_queue_iter: opening queue %r", name)
        queue = asyncio.Queue[Any](size)
        self._queues[name] = queue

        try:
            while True:
                result = await queue.get()
                yield result
        finally:
            LOGGER.debug("_queue_iter: closing queue %r", name)
            del self._queues[name]

    @TRACE
    async def run(self):
        "Run the switch's lifecycle repeatedly."
        assert self._p4client is None
        assert self._tasks is None

        self._tasks = SwitchTasks()
        self._p4client = P4Client(self._address, self._options.credentials)

        try:
            while True:
                # If the switch fails and restarts too quickly, slow it down.
                async with _throttle_failure():
                    await self._run_lifecycle()

        finally:
            self._p4client = None
            self._tasks = None

    @TRACE
    async def _run_lifecycle(self):
        "Run the switch's lifecycle once."
        assert self._tasks is not None

        self.create_task(self._run(), background=True)
        await self._tasks.wait()
        self._arbitrator.reset()

    def create_task(
        self,
        coro: Coroutine[Any, Any, _T],
        *,
        background: bool = False,
        name: str | None = None,
    ) -> asyncio.Task[_T]:
        assert self._tasks is not None

        return self._tasks.create_task(
            coro,
            switch=self,
            background=background,
            name=name,
        )

    @TRACE
    async def _run(self):
        "Main Switch task runs the stream."
        assert not self._is_channel_up
        assert self._p4client is not None

        try:
            await self._p4client.open(
                schema=self.p4info,
                complete_request=self._arbitrator.complete_request,
            )
            await self._arbitrator.handshake(self)
            await self._fetch_capabilities()
            await self._start_gnmi()
            self._channel_up()

            # Receive messages from the stream until it closes.
            await self._receive_until_closed()

        finally:
            await self._stop_gnmi()
            await self._p4client.close()
            self._channel_down()

    @TRACE
    async def _receive_until_closed(self):
        "Receive messages from stream until EOF."
        assert self._p4client is not None

        client = self._p4client

        while True:
            try:
                msg = await client.receive()
            except P4ClientError as ex:
                if not ex.status.is_election_id_used:
                    raise
                # Handle "election ID in use" error.
                await self._arbitrator.handshake(self, conflict=True)
            else:
                await self._handle_stream_message(msg)

    async def _handle_stream_message(self, msg: p4r.StreamMessageResponse):
        "Handle a P4Runtime StreamMessageResponse."

        msg_type = msg.WhichOneof("update")
        if msg_type is None:
            LOGGER.error("missing update: %r", msg)
            return

        if msg_type == "arbitration":
            await self._arbitrator.update(self, msg.arbitration)
        else:
            queue = self._queues.get(msg_type)
            if queue:
                if not queue.full():
                    obj = p4entity.decode_stream(msg, self.p4info)
                    queue.put_nowait(obj)

    async def __aenter__(self):
        "Similar to run() but provides a one-time context manager interface."
        assert self._p4client is None
        assert self._tasks is None

        self._tasks = SwitchTasks()
        self._p4client = P4Client(self._address, self._options.credentials)
        self.create_task(self._run(), background=True)

        try:
            # Wait for CHANNEL_READY event from the switch.
            await self.ee.wait_for_event(SwitchEvent.CHANNEL_READY)
        except BaseException:
            await self.__aexit__(None, None, None)
            raise

        return self

    async def __aexit__(
        self,
        _exc_type: type[BaseException] | None,
        _exc_val: BaseException | None,
        _exc_tb: TracebackType | None,
    ) -> None:
        "Similar to run() but provides a one-time context manager interface."
        assert self._p4client is not None
        assert self._tasks is not None

        self._tasks.cancel_all()
        await self._tasks.wait()
        self._arbitrator.reset()
        self._p4client = None
        self._tasks = None

    def _channel_up(self):
        assert not self._is_channel_up

        ports = " ".join(f"({port.id}){port.name}" for port in self.ports)
        LOGGER.info(
            "Channel up (is_primary=%r, election_id=%r, primary_id=%r, api_version=%s): %s",
            self.is_primary,
            self.election_id,
            self.primary_id,
            self.api_version_string,
            ports,
        )
        self._is_channel_up = True
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.CHANNEL_UP, self)

    def _channel_down(self):
        if not self._is_channel_up:
            return  # do nothing!

        LOGGER.info(
            "Channel down (is_primary=%r, election_id=%r, primary_id=%r)",
            self.is_primary,
            self.election_id,
            self.primary_id,
        )
        self._is_channel_up = False

        self.ee.emit(SwitchEvent.CHANNEL_DOWN, self)

    def _become_primary(self):
        assert self._tasks is not None

        LOGGER.info(
            "Become primary (is_primary=%r, election_id=%r, primary_id=%r)",
            self.is_primary,
            self.election_id,
            self.primary_id,
        )

        self._tasks.cancel_primary()
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.BECOME_PRIMARY, self)

    def _become_backup(self):
        assert self._tasks is not None

        LOGGER.info(
            "Become backup (is_primary=%r, election_id=%r, primary_id=%r)",
            self.is_primary,
            self.election_id,
            self.primary_id,
        )

        self._tasks.cancel_primary()
        self.create_task(self._ready())

        self.ee.emit(SwitchEvent.BECOME_BACKUP, self)

    def _channel_ready(self):
        LOGGER.info("Channel ready (is_primary=%r)", self.is_primary)

        if self._options.ready_handler:
            self.create_task(self._options.ready_handler(self))

        self.ee.emit(SwitchEvent.CHANNEL_READY, self)

    @TRACE
    async def _ready(self):
        "Prepare the pipeline."

        if self.is_primary:
            # Primary: Set up pipeline or retrieve it.
            if self.p4info.is_configured:
                await self._set_pipeline()
            else:
                await self._get_pipeline()

        elif not self.p4info.is_configured:
            # Backup: Retrieve the pipeline only if it is not configured.
            # FIXME: There is a race condition with the primary client; we
            # may receive the wrong pipeline. We may need to poll the pipeline
            # cookie when we're a backup to check for changes?
            await self._get_pipeline()

        self._channel_ready()

    @TRACE
    async def _get_pipeline(self):
        "Get the switch's P4Info."

        try:
            reply = await self._get_pipeline_config_request(
                response_type=P4ConfigResponseType.P4INFO_AND_COOKIE
            )
            if reply.config.HasField("p4info"):
                self.p4info.set_p4info(reply.config.p4info)

        except P4ClientError as ex:
            if not ex.status.is_no_pipeline_configured:
                raise

    @TRACE
    async def _set_pipeline(self):
        "Set up the pipeline."

        cookie = -1
        try:
            reply = await self._get_pipeline_config_request()
            if reply.config.HasField("cookie"):
                cookie = reply.config.cookie.cookie

        except P4ClientError as ex:
            if not ex.status.is_no_pipeline_configured:
                raise

        if cookie != self.p4info.p4cookie:
            LOGGER.debug(
                "cookie %#x does not match expected %#x", cookie, self.p4info.p4cookie
            )
            await self._set_pipeline_config_request(
                config=self.p4info.get_pipeline_config()
            )

    async def _get_pipeline_config_request(
        self,
        *,
        response_type: P4ConfigResponseType = P4ConfigResponseType.COOKIE_ONLY,
    ) -> p4r.GetForwardingPipelineConfigResponse:
        "Send a GetForwardingPipelineConfigRequest and await the response."
        assert self._p4client is not None

        return await self._p4client.request(
            p4r.GetForwardingPipelineConfigRequest(
                device_id=self.device_id,
                response_type=response_type.vt(),
            )
        )

    async def _set_pipeline_config_request(
        self,
        *,
        action: P4ConfigAction = P4ConfigAction.VERIFY_AND_COMMIT,
        config: p4r.ForwardingPipelineConfig,
    ) -> p4r.SetForwardingPipelineConfigResponse:
        "Send a SetForwardingPipelineConfigRequest and await the response."
        assert self._p4client is not None

        return await self._p4client.request(
            p4r.SetForwardingPipelineConfigRequest(
                device_id=self.device_id,
                action=action.vt(),
                config=config,
            )
        )

    async def read(self, *entities: p4entity.EntityList):
        assert self._p4client is not None

        request = p4r.ReadRequest(
            device_id=self.device_id,
            entities=p4entity.encode_entities(entities, self.p4info),
        )

        async for reply in self._p4client.request_iter(request):
            for ent in reply.entities:
                yield p4entity.decode_entity(ent, self.p4info)

    async def write(self, *entities: p4entity.UpdateList):
        assert self._p4client is not None

        msgs = p4entity.encode_updates(entities, self.p4info)

        updates: list[p4r.Update] = []
        for msg in msgs:
            if isinstance(msg, p4r.StreamMessageRequest):
                await self._p4client.send(msg)
            else:
                updates.append(msg)

        if updates:
            await self._p4client.request(
                p4r.WriteRequest(
                    device_id=self.device_id,
                    updates=updates,
                )
            )

    async def insert(self, *entities: p4entity.EntityList):
        await self._write(entities, P4UpdateType.INSERT)

    async def modify(self, *entities: p4entity.EntityList):
        await self._write(entities, P4UpdateType.MODIFY)

    async def delete(
        self,
        *entities: p4entity.EntityList,
        ignore_not_found_error: bool = False,
    ):
        """Delete the specified entities.

        It is an error to delete entities that do not exist. If
        `ignore_not_found_error` is True, this 'NOT FOUND' error will be
        suppressed.
        """
        try:
            await self._write(entities, P4UpdateType.DELETE)
        except P4ClientError as ex:
            if ignore_not_found_error and ex.status.is_not_found_only:
                return
            raise

    async def delete_all(self):
        assert self._p4client is not None

        everything = [
            p4entity.P4TableEntry(),
            p4entity.P4MulticastGroupEntry(),
            # TODO: This isn't everything...
        ]

        request = p4r.ReadRequest(
            device_id=self.device_id,
            entities=p4entity.encode_entities(everything, self.p4info),
        )

        async for reply in self._p4client.request_iter(request):
            if reply.entities:
                await self.delete(reply.entities)

        # We need to delete DigestEntry separately. Wildcard reads are not
        # supported.
        digest_entries = [
            p4entity.P4DigestEntry(digest.alias) for digest in self.p4info.digests
        ]
        if digest_entries:
            await self.delete(digest_entries, ignore_not_found_error=True)

    async def _write(self, entities: p4entity.EntityList, update_type: P4UpdateType):
        assert self._p4client is not None

        updates = [
            p4r.Update(type=update_type.vt(), entity=ent)
            for ent in p4entity.encode_entities(entities, self.p4info)
        ]
        await self._p4client.request(
            p4r.WriteRequest(
                device_id=self.device_id,
                updates=updates,
            )
        )

    async def _fetch_capabilities(self):
        "Check the P4Runtime protocol version supported by the other end."
        assert self._p4client is not None

        try:
            reply = await self._p4client.request(p4r.CapabilitiesRequest())
            self._api_version = _parse_semantic_version(reply.p4runtime_api_version)

        except P4ClientError as ex:
            if not ex.is_unimplemented:
                raise
            LOGGER.warning("CapabilitiesRequest is not implemented")

    async def _start_gnmi(self):
        assert self._gnmi_client is None
        assert self._p4client is not None

        self._gnmi_client = gNMIClient(self._address, self._options.credentials)
        await self._gnmi_client.open(channel=self._p4client.channel)

        try:
            await self._ports.subscribe(self._gnmi_client)
            self.create_task(self._ports.update(), background=True, name="_ports")

        except gNMIClientError as ex:
            if not ex.is_unimplemented:
                raise
            LOGGER.warning("gNMI is not implemented")
            await self._gnmi_client.close()
            self._gnmi_client = None

    async def _stop_gnmi(self):
        if self._gnmi_client is not None:
            self._ports.close()
            await self._gnmi_client.close()
            self._gnmi_client = None

    def __repr__(self):
        return f"Switch(name={self._name!r}, address={self._address!r})"


class SwitchEvent(str, enum.Enum):
    "Events for Switch class."

    CHANNEL_UP = "channel_up"  # (switch)
    CHANNEL_DOWN = "channel_down"  # (switch)
    CHANNEL_READY = "channel_ready"  # (switch)
    BECOME_PRIMARY = "become_primary"  # (switch)
    BECOME_BACKUP = "become_backup"  # (switch)
    PORT_UP = "port_up"  # (switch, port)
    PORT_DOWN = "port_down"  # (switch, port)
    SWITCH_DONE = "switch_done"  # (switch)


class SwitchEmitter(pyee.EventEmitter):
    "EventEmitter subclass for emitting Switch events."

    def __init__(self, switch: Switch):
        super().__init__()
        self._switch = switch

    def _emit_run(self, f: Any, args: Any, kwargs: Any):
        try:
            coro: Any = f(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.error("Error in SwitchEmitter._emit_run", exc_info=exc)
            # TODO: Have synchronous callback failures affect Switch?
        else:
            if asyncio.iscoroutine(coro):
                self._switch.create_task(coro)

    async def wait_for_event(self, event: SwitchEvent):
        "Wait for a specific event."

        ready = asyncio.Event()
        self.once(event, lambda _: ready.set())  # type: ignore
        await ready.wait()


class SwitchTasks:
    "Manage a set of related tasks."

    _tasks: set[asyncio.Task[Any]]
    _task_count: CountdownFuture

    def __init__(self):
        self._tasks = set()
        self._task_count = CountdownFuture()

    def create_task(
        self,
        coro: Coroutine[Any, Any, _T],
        *,
        switch: Switch,
        background: bool,
        name: str | None = None,
    ) -> asyncio.Task[_T]:
        "Create a new task managed by switch."
        if name is None:
            name = coro.__name__

        bg_char = "&" if background else ""
        task_name = f"fy:{switch.name}|{name}{bg_char}"

        task = asyncio.create_task(coro, name=task_name)
        self._tasks.add(task)
        self._task_count.increment()
        task.add_done_callback(self._task_done)

        return task

    def _task_done(self, done: asyncio.Task[Any]):
        "Called when a task finishes."
        self._tasks.remove(done)
        self._task_count.decrement()

        if not done.cancelled():
            ex = done.exception()
            if ex is not None:
                LOGGER.error("Switch task %r failed", done.get_name(), exc_info=ex)
                self.cancel_all()

    def cancel_all(self):
        "Cancel all tasks."
        for task in self._tasks:
            if not task.done():
                task.cancel()

    def cancel_primary(self):
        "Cancel all non-background tasks."
        for task in self._tasks:
            if not task.done() and not task.get_name().endswith("&"):
                task.cancel()

    @TRACE
    async def wait(self):
        "Wait for all tasks to finish."
        await self._task_count.wait(self.cancel_all)


@asynccontextmanager
async def _throttle_failure():
    "Used to throttle retries, if the switch's lifecycle fails too quickly."

    start_time = time.monotonic()
    yield
    end_time = time.monotonic()

    if end_time - start_time < _FAIL_MIN_TIME_SECS:
        # Handle case where we failed too quickly.
        LOGGER.warning(
            "Switch failed in less than %g seconds; waiting %g seconds to try again",
            _FAIL_MIN_TIME_SECS,
            _FAIL_THROTTLE_SECS,
        )
        await asyncio.sleep(_FAIL_THROTTLE_SECS)

    else:
        # Otherwise, always yield to other tasks here.
        await asyncio.sleep(0)


_SEMVER_REGEX = re.compile(r"^(\d+)\.(\d+)\.(\d+)")


def _parse_semantic_version(version_str: str) -> _ApiVersion:
    "Parse version string and return a 3-tuple (major, minor, patch)."

    m = _SEMVER_REGEX.match(version_str)
    if not m:
        raise ValueError(f"unexpected version string: {version_str}")

    return (int(m[1]), int(m[2]), int(m[3]))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnmi1/gnmi.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from gnmi1 import gnmi_ext_pb2 as gnmi1_dot_gnmi__ext__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10gnmi1/gnmi.proto\x12\x04gnmi\x1a\x19google/protobuf/any.proto\x1a google/protobuf/descriptor.proto\x1a\x14gnmi1/gnmi_ext.proto\"\x96\x01\n\x0cNotification\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x1a\n\x06prefix\x18\x02 \x01(\x0b\x32\n.gnmi.Path\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12\x1c\n\x06update\x18\x04 \x03(\x0b\x32\x0c.gnmi.Update\x12\x1a\n\x06\x64\x65lete\x18\x05 \x03(\x0b\x32\n.gnmi.Path\x12\x0e\n\x06\x61tomic\x18\x06 \x01(\x08\"u\n\x06Update\x12\x18\n\x04path\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0b.gnmi.ValueB\x02\x18\x01\x12\x1d\n\x03val\x18\x03 \x01(\x0b\x32\x10.gnmi.TypedValue\x12\x12\n\nduplicates\x18\x04 \x01(\r\"\x83\x03\n\nTypedValue\x12\x14\n\nstring_val\x18\x01 \x01(\tH\x00\x12\x11\n\x07int_val\x18\x02 \x01(\x03H\x00\x12\x12\n\x08uint_val\x18\x03 \x01(\x04H\x00\x12\x12\n\x08\x62ool_val\x18\x04 \x01(\x08H\x00\x12\x13\n\tbytes_val\x18\x05 \x01(\x0cH\x00\x12\x17\n\tfloat_val\x18\x06 \x01(\x02\x42\x02\x18\x01H\x00\x12\x14\n\ndouble_val\x18\x0e \x01(\x01H\x00\x12*\n\x0b\x64\x65\x63imal_val\x18\x07 \x01(\x0b\x32\x0f.gnmi.Decimal64B\x02\x18\x01H\x00\x12)\n\x0cleaflist_val\x18\x08 \x01(\x0b\x32\x11.gnmi.ScalarArrayH\x00\x12\'\n\x07\x61ny_val\x18\t \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x12\x12\n\x08json_val\x18\n \x01(\x0cH\x00\x12\x17\n\rjson_ietf_val\x18\x0b \x01(\x0cH\x00\x12\x13\n\tascii_val\x18\x0c \x01(\tH\x00\x12\x15\n\x0bproto_bytes\x18\r \x01(\x0cH\x00\x42\x07\n\x05value\"Y\n\x04Path\x12\x13\n\x07\x65lement\x18\x01 \x03(\tB\x02\x18\x01\x12\x0e\n\x06origin\x18\x02 \x01(\t\x12\x1c\n\x04\x65lem\x18\x03 \x03(\x0b\x32\x0e.gnmi.PathElem\x12\x0e\n\x06target\x18\x04 \x01(\t\"j\n\x08PathElem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x03key\x18\x02 \x03(\x0b\x32\x17.gnmi.PathElem.KeyEntry\x1a*\n\x08KeyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"8\n\x05Value\x12\r\n\x05value\x18\x01 \x01(\x0c\x12\x1c\n\x04type\x18\x02 \x01(\x0e\x32\x0e.gnmi.Encoding:\x02\x18\x01\"N\n\x05\x45rror\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x18\x01\".\n\tDecimal64\x12\x0e\n\x06\x64igits\x18\x01 \x01(\x03\x12\x11\n\tprecision\x18\x02 \x01(\r\"0\n\x0bScalarArray\x12!\n\x07\x65lement\x18\x01 \x03(\x0b\x32\x10.gnmi.TypedValue\"\xb2\x01\n\x10SubscribeRequest\x12+\n\tsubscribe\x18\x01 \x01(\x0b\x32\x16.gnmi.SubscriptionListH\x00\x12\x1a\n\x04poll\x18\x03 \x01(\x0b\x32\n.gnmi.PollH\x00\x12\"\n\x07\x61liases\x18\x04 \x01(\x0b\x32\x0f.gnmi.AliasListH\x00\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.ExtensionB\t\n\x07request\"\x06\n\x04Poll\"\xa8\x01\n\x11SubscribeResponse\x12$\n\x06update\x18\x01 \x01(\x0b\x32\x12.gnmi.NotificationH\x00\x12\x17\n\rsync_response\x18\x03 \x01(\x08H\x00\x12 \n\x05\x65rror\x18\x04 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01H\x00\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.ExtensionB\n\n\x08response\"\xd7\x02\n\x10SubscriptionList\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12(\n\x0csubscription\x18\x02 \x03(\x0b\x32\x12.gnmi.Subscription\x12\x13\n\x0buse_aliases\x18\x03 \x01(\x08\x12\x1d\n\x03qos\x18\x04 \x01(\x0b\x32\x10.gnmi.QOSMarking\x12)\n\x04mode\x18\x05 \x01(\x0e\x32\x1b.gnmi.SubscriptionList.Mode\x12\x19\n\x11\x61llow_aggregation\x18\x06 \x01(\x08\x12#\n\nuse_models\x18\x07 \x03(\x0b\x32\x0f.gnmi.ModelData\x12 \n\x08\x65ncoding\x18\x08 \x01(\x0e\x32\x0e.gnmi.Encoding\x12\x14\n\x0cupdates_only\x18\t \x01(\x08\"&\n\x04Mode\x12\n\n\x06STREAM\x10\x00\x12\x08\n\x04ONCE\x10\x01\x12\x08\n\x04POLL\x10\x02\"\x9f\x01\n\x0cSubscription\x12\x18\n\x04path\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12$\n\x04mode\x18\x02 \x01(\x0e\x32\x16.gnmi.SubscriptionMode\x12\x17\n\x0fsample_interval\x18\x03 \x01(\x04\x12\x1a\n\x12suppress_redundant\x18\x04 \x01(\x08\x12\x1a\n\x12heartbeat_interval\x18\x05 \x01(\x04\"\x1d\n\nQOSMarking\x12\x0f\n\x07marking\x18\x01 \x01(\r\"0\n\x05\x41lias\x12\x18\n\x04path\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"\'\n\tAliasList\x12\x1a\n\x05\x61lias\x18\x01 \x03(\x0b\x32\x0b.gnmi.Alias\"\xa9\x01\n\nSetRequest\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x1a\n\x06\x64\x65lete\x18\x02 \x03(\x0b\x32\n.gnmi.Path\x12\x1d\n\x07replace\x18\x03 \x03(\x0b\x32\x0c.gnmi.Update\x12\x1c\n\x06update\x18\x04 \x03(\x0b\x32\x0c.gnmi.Update\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xac\x01\n\x0bSetResponse\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12$\n\x08response\x18\x02 \x03(\x0b\x32\x12.gnmi.UpdateResult\x12 \n\x07message\x18\x03 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12\x11\n\ttimestamp\x18\x04 \x01(\x03\x12&\n\textension\x18\x05 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xca\x01\n\x0cUpdateResult\x12\x15\n\ttimestamp\x18\x01 \x01(\x03\x42\x02\x18\x01\x12\x18\n\x04path\x18\x02 \x01(\x0b\x32\n.gnmi.Path\x12 \n\x07message\x18\x03 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12(\n\x02op\x18\x04 \x01(\x0e\x32\x1c.gnmi.UpdateResult.Operation\"=\n\tOperation\x12\x0b\n\x07INVALID\x10\x00\x12\n\n\x06\x44\x45LETE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02\x12\n\n\x06UPDATE\x10\x03\"\x97\x02\n\nGetRequest\x12\x1a\n\x06prefix\x18\x01 \x01(\x0b\x32\n.gnmi.Path\x12\x18\n\x04path\x18\x02 \x03(\x0b\x32\n.gnmi.Path\x12\'\n\x04type\x18\x03 \x01(\x0e\x32\x19.gnmi.GetRequest.DataType\x12 \n\x08\x65ncoding\x18\x05 \x01(\x0e\x32\x0e.gnmi.Encoding\x12#\n\nuse_models\x18\x06 \x03(\x0b\x32\x0f.gnmi.ModelData\x12&\n\textension\x18\x07 \x03(\x0b\x32\x13.gnmi_ext.Extension\";\n\x08\x44\x61taType\x12\x07\n\x03\x41LL\x10\x00\x12\n\n\x06\x43ONFIG\x10\x01\x12\t\n\x05STATE\x10\x02\x12\x0f\n\x0bOPERATIONAL\x10\x03\"\x7f\n\x0bGetResponse\x12(\n\x0cnotification\x18\x01 \x03(\x0b\x32\x12.gnmi.Notification\x12\x1e\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x0b.gnmi.ErrorB\x02\x18\x01\x12&\n\textension\x18\x03 \x03(\x0b\x32\x13.gnmi_ext.Extension\";\n\x11\x43\x61pabilityRequest\x12&\n\textension\x18\x01 \x03(\x0b\x32\x13.gnmi_ext.Extension\"\xaa\x01\n\x12\x43\x61pabilityResponse\x12)\n\x10supported_models\x18\x01 \x03(\x0b\x32\x0f.gnmi.ModelData\x12+\n\x13supported_encodings\x18\x02 \x03(\x0e\x32\x0e.gnmi.Encoding\x12\x14\n\x0cgNMI_version\x18\x03 \x01(\t\x12&\n\textension\x18\x04 \x03(\x0b\x32\x13.gnmi_ext.Extension\"@\n\tModelData\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0corganization\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t*D\n\x08\x45ncoding\x12\x08\n\x04JSON\x10\x00\x12\t\n\x05\x42YTES\x10\x01\x12\t\n\x05PROTO\x10\x02\x12\t\n\x05\x41SCII\x10\x03\x12\r\n\tJSON_IETF\x10\x04*A\n\x10SubscriptionMode\x12\x12\n\x0eTARGET_DEFINED\x10\x00\x12\r\n\tON_CHANGE\x10\x01\x12\n\n\x06SAMPLE\x10\x02\x32\xe3\x01\n\x04gNMI\x12\x41\n\x0c\x43\x61pabilities\x12\x17.gnmi.CapabilityRequest\x1a\x18.gnmi.CapabilityResponse\x12*\n\x03Get\x12\x10.gnmi.GetRequest\x1a\x11.gnmi.GetResponse\x12*\n\x03Set\x12\x10.gnmi.SetRequest\x1a\x11.gnmi.SetResponse\x12@\n\tSubscribe\x12\x16.gnmi.SubscribeRequest\x1a\x17.gnmi.SubscribeResponse(\x01\x30\x01:3\n\x0cgnmi_service\x12\x1c.google.protobuf.FileOptions\x18\xe9\x07 \x01(\tBS\n\x15\x63om.github.gnmi.protoB\tGnmiProtoP\x01Z%github.com/openconfig/gnmi/proto/gnmi\xca>\x05\x30.8.0b\x06proto3')

_ENCODING = DESCRIPTOR.enum_types_by_name['Encoding']
Encoding = enum_type_wrapper.EnumTypeWrapper(_ENCODING)
_SUBSCRIPTIONMODE = DESCRIPTOR.enum_types_by_name['SubscriptionMode']
SubscriptionMode = enum_type_wrapper.EnumTypeWrapper(_SUBSCRIPTIONMODE)
JSON = 0
BYTES = 1
PROTO = 2
ASCII = 3
JSON_IETF = 4
TARGET_DEFINED = 0
ON_CHANGE = 1
SAMPLE = 2

GNMI_SERVICE_FIELD_NUMBER = 1001
gnmi_service = DESCRIPTOR.extensions_by_name['gnmi_service']

_NOTIFICATION = DESCRIPTOR.message_types_by_name['Notification']
_UPDATE = DESCRIPTOR.message_types_by_name['Update']
_TYPEDVALUE = DESCRIPTOR.message_types_by_name['TypedValue']
_PATH = DESCRIPTOR.message_types_by_name['Path']
_PATHELEM = DESCRIPTOR.message_types_by_name['PathElem']
_PATHELEM_KEYENTRY = _PATHELEM.nested_types_by_name['KeyEntry']
_VALUE = DESCRIPTOR.message_types_by_name['Value']
_ERROR = DESCRIPTOR.message_types_by_name['Error']
_DECIMAL64 = DESCRIPTOR.message_types_by_name['Decimal64']
_SCALARARRAY = DESCRIPTOR.message_types_by_name['ScalarArray']
_SUBSCRIBEREQUEST = DESCRIPTOR.message_types_by_name['SubscribeRequest']
_POLL = DESCRIPTOR.message_types_by_name['Poll']
_SUBSCRIBERESPONSE = DESCRIPTOR.message_types_by_name['SubscribeResponse']
_SUBSCRIPTIONLIST = DESCRIPTOR.message_types_by_name['SubscriptionList']
_SUBSCRIPTION = DESCRIPTOR.message_types_by_name['Subscription']
_QOSMARKING = DESCRIPTOR.message_types_by_name['QOSMarking']
_ALIAS = DESCRIPTOR.message_types_by_name['Alias']
_ALIASLIST = DESCRIPTOR.message_types_by_name['AliasList']
_SETREQUEST = DESCRIPTOR.message_types_by_name['SetRequest']
_SETRESPONSE = DESCRIPTOR.message_types_by_name['SetResponse']
_UPDATERESULT = DESCRIPTOR.message_types_by_name['UpdateResult']
_GETREQUEST = DESCRIPTOR.message_types_by_name['GetRequest']
_GETRESPONSE = DESCRIPTOR.message_types_by_name['GetResponse']
_CAPABILITYREQUEST = DESCRIPTOR.message_types_by_name['CapabilityRequest']
_CAPABILITYRESPONSE = DESCRIPTOR.message_types_by_name['CapabilityResponse']
_MODELDATA = DESCRIPTOR.message_types_by_name['ModelData']
_SUBSCRIPTIONLIST_MODE = _SUBSCRIPTIONLIST.enum_types_by_name['Mode']
_UPDATERESULT_OPERATION = _UPDATERESULT.enum_types_by_name['Operation']
_GETREQUEST_DATATYPE = _GETREQUEST.enum_types_by_name['DataType']
Notification = _reflection.GeneratedProtocolMessageType('Notification', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFICATION,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Notification)
  })
_sym_db.RegisterMessage(Notification)

Update = _reflection.GeneratedProtocolMessageType('Update', (_message.Message,), {
  'DESCRIPTOR' : _UPDATE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Update)
  })
_sym_db.RegisterMessage(Update)

TypedValue = _reflection.GeneratedProtocolMessageType('TypedValue', (_message.Message,), {
  'DESCRIPTOR' : _TYPEDVALUE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.TypedValue)
  })
_sym_db.RegisterMessage(TypedValue)

Path = _reflection.GeneratedProtocolMessageType('Path', (_message.Message,), {
  'DESCRIPTOR' : _PATH,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Path)
  })
_sym_db.RegisterMessage(Path)

PathElem = _reflection.GeneratedProtocolMessageType('PathElem', (_message.Message,), {

  'KeyEntry' : _reflection.GeneratedProtocolMessageType('KeyEntry', (_message.Message,), {
    'DESCRIPTOR' : _PATHELEM_KEYENTRY,
    '__module__' : 'gnmi1.gnmi_pb2'
    # @@protoc_insertion_point(class_scope:gnmi.PathElem.KeyEntry)
    })
  ,
  'DESCRIPTOR' : _PATHELEM,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.PathElem)
  })
_sym_db.RegisterMessage(PathElem)
_sym_db.RegisterMessage(PathElem.KeyEntry)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Value)
  })
_sym_db.RegisterMessage(Value)

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
  'DESCRIPTOR' : _ERROR,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Error)
  })
_sym_db.RegisterMessage(Error)

Decimal64 = _reflection.GeneratedProtocolMessageType('Decimal64', (_message.Message,), {
  'DESCRIPTOR' : _DECIMAL64,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Decimal64)
  })
_sym_db.RegisterMessage(Decimal64)

ScalarArray = _reflection.GeneratedProtocolMessageType('ScalarArray', (_message.Message,), {
  'DESCRIPTOR' : _SCALARARRAY,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.ScalarArray)
  })
_sym_db.RegisterMessage(ScalarArray)

SubscribeRequest = _reflection.GeneratedProtocolMessageType('SubscribeRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIBEREQUEST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.SubscribeRequest)
  })
_sym_db.RegisterMessage(SubscribeRequest)

Poll = _reflection.GeneratedProtocolMessageType('Poll', (_message.Message,), {
  'DESCRIPTOR' : _POLL,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Poll)
  })
_sym_db.RegisterMessage(Poll)

SubscribeResponse = _reflection.GeneratedProtocolMessageType('SubscribeResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIBERESPONSE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.SubscribeResponse)
  })
_sym_db.RegisterMessage(SubscribeResponse)

SubscriptionList = _reflection.GeneratedProtocolMessageType('SubscriptionList', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONLIST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.SubscriptionList)
  })
_sym_db.RegisterMessage(SubscriptionList)

Subscription = _reflection.GeneratedProtocolMessageType('Subscription', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTION,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Subscription)
  })
_sym_db.RegisterMessage(Subscription)

QOSMarking = _reflection.GeneratedProtocolMessageType('QOSMarking', (_message.Message,), {
  'DESCRIPTOR' : _QOSMARKING,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.QOSMarking)
  })
_sym_db.RegisterMessage(QOSMarking)

Alias = _reflection.GeneratedProtocolMessageType('Alias', (_message.Message,), {
  'DESCRIPTOR' : _ALIAS,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.Alias)
  })
_sym_db.RegisterMessage(Alias)

AliasList = _reflection.GeneratedProtocolMessageType('AliasList', (_message.Message,), {
  'DESCRIPTOR' : _ALIASLIST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.AliasList)
  })
_sym_db.RegisterMessage(AliasList)

SetRequest = _reflection.GeneratedProtocolMessageType('SetRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETREQUEST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.SetRequest)
  })
_sym_db.RegisterMessage(SetRequest)

SetResponse = _reflection.GeneratedProtocolMessageType('SetResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETRESPONSE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.SetResponse)
  })
_sym_db.RegisterMessage(SetResponse)

UpdateResult = _reflection.GeneratedProtocolMessageType('UpdateResult', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERESULT,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.UpdateResult)
  })
_sym_db.RegisterMessage(UpdateResult)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREQUEST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.GetRequest)
  })
_sym_db.RegisterMessage(GetRequest)

GetResponse = _reflection.GeneratedProtocolMessageType('GetResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETRESPONSE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.GetResponse)
  })
_sym_db.RegisterMessage(GetResponse)

CapabilityRequest = _reflection.GeneratedProtocolMessageType('CapabilityRequest', (_message.Message,), {
  'DESCRIPTOR' : _CAPABILITYREQUEST,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.CapabilityRequest)
  })
_sym_db.RegisterMessage(CapabilityRequest)

CapabilityResponse = _reflection.GeneratedProtocolMessageType('CapabilityResponse', (_message.Message,), {
  'DESCRIPTOR' : _CAPABILITYRESPONSE,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.CapabilityResponse)
  })
_sym_db.RegisterMessage(CapabilityResponse)

ModelData = _reflection.GeneratedProtocolMessageType('ModelData', (_message.Message,), {
  'DESCRIPTOR' : _MODELDATA,
  '__module__' : 'gnmi1.gnmi_pb2'
  # @@protoc_insertion_point(class_scope:gnmi.ModelData)
  })
_sym_db.RegisterMessage(ModelData)

_GNMI = DESCRIPTOR.services_by_name['gNMI']
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(gnmi_service)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\025com.github.gnmi.protoB\tGnmiProtoP\001Z%github.com/openconfig/gnmi/proto/gnmi\312>\0050.8.0'
  _UPDATE.fields_by_name['value']._options = None
  _UPDATE.fields_by_name['value']._serialized_options = b'\030\001'
  _TYPEDVALUE.fields_by_name['float_val']._options = None
  _TYPEDVALUE.fields_by_name['float_val']._serialized_options = b'\030\001'
  _TYPEDVALUE.fields_by_name['decimal_val']._options = None
  _TYPEDVALUE.fields_by_name['decimal_val']._serialized_options = b'\030\001'
  _PATH.fields_by_name['element']._options = None
  _PATH.fields_by_name['element']._serialized_options = b'\030\001'
  _PATHELEM_KEYENTRY._options = None
  _PATHELEM_KEYENTRY._serialized_options = b'8\001'
  _VALUE._options = None
  _VALUE._serialized_options = b'\030\001'
  _ERROR._options = None
  _ERROR._serialized_options = b'\030\001'
  _SUBSCRIBERESPONSE.fields_by_name['error']._options = None
  _SUBSCRIBERESPONSE.fields_by_name['error']._serialized_options = b'\030\001'
  _SETRESPONSE.fields_by_name['message']._options = None
  _SETRESPONSE.fields_by_name['message']._serialized_options = b'\030\001'
  _UPDATERESULT.fields_by_name['timestamp']._options = None
  _UPDATERESULT.fields_by_name['timestamp']._serialized_options = b'\030\001'
  _UPDATERESULT.fields_by_name['message']._options = None
  _UPDATERESULT.fields_by_name['message']._serialized_options = b'\030\001'
  _GETRESPONSE.fields_by_name['error']._options = None
  _GETRESPONSE.fields_by_name['error']._serialized_options = b'\030\001'
  _ENCODING._serialized_start=3459
  _ENCODING._serialized_end=3527
  _SUBSCRIPTIONMODE._serialized_start=3529
  _SUBSCRIPTIONMODE._serialized_end=3594
  _NOTIFICATION._serialized_start=110
  _NOTIFICATION._serialized_end=260
  _UPDATE._serialized_start=262
  _UPDATE._serialized_end=379
  _TYPEDVALUE._serialized_start=382
  _TYPEDVALUE._serialized_end=769
  _PATH._serialized_start=771
  _PATH._serialized_end=860
  _PATHELEM._serialized_start=862
  _PATHELEM._serialized_end=968
  _PATHELEM_KEYENTRY._serialized_start=926
  _PATHELEM_KEYENTRY._serialized_end=968
  _VALUE._serialized_start=970
  _VALUE._serialized_end=1026
  _ERROR._serialized_start=1028
  _ERROR._serialized_end=1106
  _DECIMAL64._serialized_start=1108
  _DECIMAL64._serialized_end=1154
  _SCALARARRAY._serialized_start=1156
  _SCALARARRAY._serialized_end=1204
  _SUBSCRIBEREQUEST._serialized_start=1207
  _SUBSCRIBEREQUEST._serialized_end=1385
  _POLL._serialized_start=1387
  _POLL._serialized_end=1393
  _SUBSCRIBERESPONSE._serialized_start=1396
  _SUBSCRIBERESPONSE._serialized_end=1564
  _SUBSCRIPTIONLIST._serialized_start=1567
  _SUBSCRIPTIONLIST._serialized_end=1910
  _SUBSCRIPTIONLIST_MODE._serialized_start=1872
  _SUBSCRIPTIONLIST_MODE._serialized_end=1910
  _SUBSCRIPTION._serialized_start=1913
  _SUBSCRIPTION._serialized_end=2072
  _QOSMARKING._serialized_start=2074
  _QOSMARKING._serialized_end=2103
  _ALIAS._serialized_start=2105
  _ALIAS._serialized_end=2153
  _ALIASLIST._serialized_start=2155
  _ALIASLIST._serialized_end=2194
  _SETREQUEST._serialized_start=2197
  _SETREQUEST._serialized_end=2366
  _SETRESPONSE._serialized_start=2369
  _SETRESPONSE._serialized_end=2541
  _UPDATERESULT._serialized_start=2544
  _UPDATERESULT._serialized_end=2746
  _UPDATERESULT_OPERATION._serialized_start=2685
  _UPDATERESULT_OPERATION._serialized_end=2746
  _GETREQUEST._serialized_start=2749
  _GETREQUEST._serialized_end=3028
  _GETREQUEST_DATATYPE._serialized_start=2969
  _GETREQUEST_DATATYPE._serialized_end=3028
  _GETRESPONSE._serialized_start=3030
  _GETRESPONSE._serialized_end=3157
  _CAPABILITYREQUEST._serialized_start=3159
  _CAPABILITYREQUEST._serialized_end=3218
  _CAPABILITYRESPONSE._serialized_start=3221
  _CAPABILITYRESPONSE._serialized_end=3391
  _MODELDATA._serialized_start=3393
  _MODELDATA._serialized_end=3457
  _GNMI._serialized_start=3597
  _GNMI._serialized_end=3824
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import gnmi1.gnmi_pb2
import grpc
import typing

class gNMIStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Capabilities: grpc.UnaryUnaryMultiCallable[
        gnmi1.gnmi_pb2.CapabilityRequest,
        gnmi1.gnmi_pb2.CapabilityResponse]
    """Capabilities allows the client to retrieve the set of capabilities that
    is supported by the target. This allows the target to validate the
    service version that is implemented and retrieve the set of models that
    the target supports. The models can then be specified in subsequent RPCs
    to restrict the set of data that is utilized.
    Reference: gNMI Specification Section 3.2
    """

    Get: grpc.UnaryUnaryMultiCallable[
        gnmi1.gnmi_pb2.GetRequest,
        gnmi1.gnmi_pb2.GetResponse]
    """Retrieve a snapshot of data from the target. A Get RPC requests that the
    target snapshots a subset of the data tree as specified by the paths
    included in the message and serializes this to be returned to the
    client using the specified encoding.
    Reference: gNMI Specification Section 3.3
    """

    Set: grpc.UnaryUnaryMultiCallable[
        gnmi1.gnmi_pb2.SetRequest,
        gnmi1.gnmi_pb2.SetResponse]
    """Set allows the client to modify the state of data on the target. The
    paths to modified along with the new values that the client wishes
    to set the value to.
    Reference: gNMI Specification Section 3.4
    """

    Subscribe: grpc.StreamStreamMultiCallable[
        gnmi1.gnmi_pb2.SubscribeRequest,
        gnmi1.gnmi_pb2.SubscribeResponse]
    """Subscribe allows a client to request the target to send it values
    of particular paths within the data tree. These values may be streamed
    at a particular cadence (STREAM), sent one off on a long-lived channel
    (POLL), or sent as a one-off retrieval (ONCE).
    Reference: gNMI Specification Section 3.5
    """


class gNMIServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Capabilities(self,
        request: gnmi1.gnmi_pb2.CapabilityRequest,
        context: grpc.ServicerContext,
    ) -> gnmi1.gnmi_pb2.CapabilityResponse:
        """Capabilities allows the client to retrieve the set of capabilities that
        is supported by the target. This allows the target to validate the
        service version that is implemented and retrieve the set of models that
        the target supports. The models can then be specified in subsequent RPCs
        to restrict the set of data that is utilized.
        Reference: gNMI Specification Section 3.2
        """
        pass

    @abc.abstractmethod
    def Get(self,
        request: gnmi1.gnmi_pb2.GetRequest,
        context: grpc.ServicerContext,
    ) -> gnmi1.gnmi_pb2.GetResponse:
        """Retrieve a snapshot of data from the target. A Get RPC requests that the
        target snapshots a subset of the data tree as specified by the paths
        included in the message and serializes this to be returned to the
        client using the specified encoding.
        Reference: gNMI Specification Section 3.3
        """
        pass

    @abc.abstractmethod
    def Set(self,
        request: gnmi1.gnmi_pb2.SetRequest,
        context: grpc.ServicerContext,
    ) -> gnmi1.gnmi_pb2.SetResponse:
        """Set allows the client to modify the state of data on the target. The
        paths to modified along with the new values that the client wishes
        to set the value to.
        Reference: gNMI Specification Section 3.4
        """
        pass

    @abc.abstractmethod
    def Subscribe(self,
        request_iterator: typing.Iterator[gnmi1.gnmi_pb2.SubscribeRequest],
        context: grpc.ServicerContext,
    ) -> typing.Iterator[gnmi1.gnmi_pb2.SubscribeResponse]:
        """Subscribe allows a client to request the target to send it values
        of particular paths within the data tree. These values may be streamed
        at a particular cadence (STREAM), sent one off on a long-lived channel
        (POLL), or sent as a one-off retrieval (ONCE).
        Reference: gNMI Specification Section 3.5
        """
        pass


def add_gNMIServicer_to_server(servicer: gNMIServicer, server: grpc.Server) -> None: ...



"Implements the Controller class."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Any, Iterable

from finsy.log import LOGGER, TRACE
from finsy.switch import Switch, SwitchEvent
from finsy.util import CountdownFuture, wait_for_cancel


class Controller:
    """A Controller is a collection of switches.

    Each switch MUST have a unique name.
    """

    _name: str  # used to name control_task
    _switches: dict[str, Switch]
    _removed: set[Switch]
    _task_count: CountdownFuture

    control_task: asyncio.Task[Any] | None = None
    "Keep track of controller's main task."

    def __init__(self, switches: Iterable[Switch], *, name="controller"):
        self._name = name
        self._switches = {}
        self._removed = set()
        self._task_count = CountdownFuture()

        for switch in switches:
            if switch.name in self._switches:
                raise ValueError(f"Switch named {switch.name!r} already exists")
            self._switches[switch.name] = switch

    @property
    def running(self) -> bool:
        "True if Controller is running."
        return self.control_task is not None

    @TRACE
    async def run(self):
        "Run the controller."
        assert not self.running, "Controller.run() is not re-entrant"

        self.control_task = asyncio.current_task()
        assert self.control_task is not None
        self.control_task.set_name(f"fy:{self._name}")

        try:
            # Start each switch running.
            for switch in self:
                self._start_switch(switch)

            # Run forever.
            await wait_for_cancel()

            # Stop all the switches.
            for switch in self:
                self._stop_switch(switch)

            # Wait for switch tasks to finish.
            await self._task_count.wait()

        finally:
            self.control_task = None

    def stop(self):
        "Stop the controller if it is running."

        if self.control_task is not None:
            self.control_task.cancel()

    def add(self, switch: Switch) -> None:
        """Add a switch to the controller.

        If the controller is running, tell the switch to start.
        """
        if switch.name in self._switches:
            raise ValueError(f"Switch named {switch.name!r} already exists")

        assert switch.control_task is None

        self._switches[switch.name] = switch
        if self.running:
            self._start_switch(switch)

    def remove(self, switch: Switch) -> None:
        """Remove a switch from the controller.

        If the controller is running, tell the switch to stop and schedule it
        for removal when it fully stops.
        """
        name = switch.name
        if self._switches.get(name, None) is not switch:
            raise ValueError(f"Switch named {name!r} not found")

        del self._switches[name]

        if self.running:
            self._stop_switch(switch)
            self._removed.add(switch)
            switch.ee.once(SwitchEvent.SWITCH_DONE, self._removed.discard)

    def _start_switch(self, switch: Switch):
        "Start the switch's control task."
        LOGGER.debug("Controller._start_switch: %r", switch)
        assert switch.control_task is None

        task = asyncio.create_task(switch.run(), name=f"fy:{switch.name}")
        switch.control_task = task
        self._task_count.increment()

        def _switch_done(done):
            switch.control_task = None
            switch.ee.emit(SwitchEvent.SWITCH_DONE)
            self._task_count.decrement()

            if not done.cancelled():
                ex = done.exception()
                if ex is not None:
                    LOGGER.error(
                        "Controller task %r failed",
                        done.get_name(),
                        exc_info=ex,
                    )

        task.add_done_callback(_switch_done)

    def _stop_switch(self, switch: Switch):
        "Stop the switch's control task."
        LOGGER.debug("Controller._stop_switch: %r", switch)

        if switch.control_task is not None:
            switch.control_task.cancel()

    def __len__(self):
        "Return switch count."
        return len(self._switches)

    def __iter__(self):
        "Iterate over switches."
        return iter(self._switches.values())

    def __getitem__(self, name):
        "Retrieve switch by name."
        return self._switches[name]
